#include "../utils/Helper.h"
#include "../blops6/blackops6.h"
#include "../maps.h"
#include <thread>
#include <cmath>

void CycleMaps(DXRender& Render, int& index, mapInfo& currentMapInfo,
    const std::unordered_map<std::string, mapInfo>& mapBounds,
    const std::vector<std::string>& mapKeys) {

    index = (index + 1) % mapKeys.size();
    const std::string& currentMap = mapKeys[index];
    currentMapInfo = mapBounds.at(currentMap);

    Render.LoadTexture(currentMap);
    Render.Calculate();
}

void Render(Game& game, DXRender& Render, DXOverlay& Overlay, mapInfo& mapInfo, Player& localPlayer) {
    MSG m;
    int index = 0;
    int pos;

    uint64_t color_red = 0xFFFF0000;
    uint64_t color_green = 0xFF00FF00;
    uint64_t color_white = 0xFFFFFFFE;
    uint64_t color_magenta = 0xFFFF00FF;
    uint64_t color_1 = color_green;
    uint64_t color_2 = color_red;
    uint64_t swap;
    uint64_t color_use = color_magenta;
    bool names = true;

    std::vector<std::string> mapKeys;
    for (const auto& entry : mapBounds) {
        mapKeys.push_back(entry.first);
    }

    Render.GetViewport();
    CycleMaps(Render, index, mapInfo, mapBounds, mapKeys);

    do {
        if (PeekMessage(&m, Overlay.GetOverlay(), NULL, NULL, PM_REMOVE)) {
            TranslateMessage(&m);
            DispatchMessage(&m);
        }

        Render.BeginRender();

        // map list
        pos = 20;
        for (int i = 0; i < mapKeys.size(); i++) {
            if (i == index) {
                Render.DrawTextX(mapKeys[i], 20, pos, color_red, false, false);
            }
            else {
                Render.DrawTextX(mapKeys[i], 20, pos, color_white, false, false);
            }
            pos += 20;
        }
        pos += 20;
        Render.DrawTextX("[F3] Change map", 20, pos, color_white, false, false); pos += 20;
        Render.DrawTextX("[F4] Swap team color", 20, pos, color_white, false, false); pos += 20;
        Render.DrawTextX("[F5] Enable/Disable names", 20, pos, color_white, false, false); pos += 20;
        Render.DrawTextX("[F6] Retry localPlayer", 20, pos, color_white, false, false);

        // MAP CHANGE [F3]
        if (GetAsyncKeyState(VK_F3) & 1) {
            CycleMaps(Render, index, mapInfo, mapBounds, mapKeys);
        }
        // COLOR SWAP [F4]
        else if (GetAsyncKeyState(VK_F4) & 1) {
            swap = color_1;
            color_1 = color_2;
            color_2 = swap;
        } 
        // DISABLE/ENABLE NAMES [F5]
        else if (GetAsyncKeyState(VK_F5) & 1) {
            if (names) {
                names = false;
            }
            else {
                names = true;
            }
        }
        // RETRY LOCAL PLAYER [F6]
        else if (GetAsyncKeyState(VK_F6) & 1) {
            localPlayer.get_index(game.clientInfo);
            game.Players.clear();
            game.populatePlayers(game.hand, localPlayer);
            game.gatherPlayers(game.hand);
        }

        // check if in game
        if (game.InGame == 1) {
            Render.DrawTextX("Not in game!", Render.w / 2, Render.h / 2, color_red);
            Render.EndRender();
            continue;
        }

        // render map
        Render.RenderMap();

        // render player loop
        if (game.Players.size() > 0) {

            for (int i = 0; i < game.Players.size(); i++) {

                if (game.Players[i].dead > 0 || !game.Players[i].valid) {
                    continue;
                }

                if (game.Players[i].local) {
                    color_use = color_magenta;

                } else if (game.Players[i].teamId == game.team1) {
                    color_use = color_1;

                } else {
                    color_use = color_2;
                }

                Render.DrawFilledRectangle(game.Players[i].x - 5, game.Players[i].y, 10, 10, color_use);
                Render.DrawLine(game.Players[i].x, game.Players[i].y, game.Players[i].line.x, game.Players[i].line.y, color_use);

                // remove && and afterif you want localplayer name
                if (names && !game.Players[i].local) {
                    Render.DrawTextX(game.Players[i].name, game.Players[i].x, game.Players[i].y - 10, color_white, true, false, 1);
                }

            }

        }

        Render.EndRender();

    } while (m.message != WM_QUIT);
}

int main() {
	DMAController DMA;
	bool gameRunning = DMA.GetProcessInfo((LPSTR)"cod.exe");
    if (!gameRunning) {
        printf("killing program\n");
        exit(0);
    }

    float n_x, n_y;
    mapInfo mapInfo = { 0, 0, 0, 0 };
    Game game;
    bool result;
    float rad;

    int line_len = 25;
    double pi = 3.141592653;

    VMMDLL_SCATTER_HANDLE handle = CreateScatterHandle();
    DXOverlay Overlay;
    DXRender _Render = Overlay.CreateRender();
    Player localPlayer;

    std::thread RenderThread(Render, std::ref(game), std::ref(_Render), std::ref(Overlay), std::ref(mapInfo), std::ref(localPlayer));

    while (true) {
        game.scatter_ingame(handle);
        result = ExecuteScatterRead(handle);
        if (!result) {
            printf("scatter read failed\n");
        }

        if (game.InGame > 1) {
            std::this_thread::sleep_for(std::chrono::milliseconds(10000));
            game.clientInfo = decrypt_client_info();

            if (game.clientInfo) {
                // Collect game info
                game.clientBase = decrypt_client_base(game.clientInfo);
                game.refDefPtr = decrypt_refdef->retrieve_ref_def();
                game.name_ptr = ReadValue<uint64_t>(Addresses::base + xbox::name_array);
                game.name_start = game.name_ptr + xbox::name_array_pos;

                // Collect localPlayer
                localPlayer.get_index(game.clientInfo);
                localPlayer.Player = game.clientBase + (localPlayer.index * xbox::player::size);
                localPlayer.info(handle);
                AddScatterRead(handle, game.refDefPtr, &decrypt_refdef->ref_def_nn, sizeof(ref_def_t));
                result = ExecuteScatterRead(handle);
                if (!result) {
                    printf("localPlayer, refdef\n");
                }

                // refDef and team set
                ref_def_t& ref_def = decrypt_refdef->ref_def_nn;
                game.team1 = localPlayer.teamId;

                // populate Player list
                game.populatePlayers(handle, localPlayer);

                // gather info and names
                game.gatherPlayers(handle);

                int count = 0;
                while (true) {
                    count++;

                    // update player position
                    for (int i = 0; i < game.Players.size(); i++) {
                        if (game.Players[i].valid) {
                            game.Players[i].updatePlayer(handle);
                        }
                    }
                    
                    // extremely robust and mature leaver check
                    if (count == 1000) {
                        if (!game.checkValid(handle)) {
                            game.gatherPlayers(handle);
                        }
                        count = 0;
                    }
                    
                    // check if in game
                    game.scatter_ingame(handle);
                    result = ExecuteScatterRead(handle);
                    if (!result) {
                        printf("in game check failed\n");
                    }

                    // clear list and wait for game start
                    if (game.InGame == 1) {
                        game.Players.clear();
                        break;
                    }

                    for (int i = 0; i < game.Players.size(); i++) {
                        if (game.Players[i].valid) {

                            std::lock_guard<std::mutex> lock(mutexlock);

                            // Debug shit
                            //printf("Player %d, %llx | dead %d\n ", game.Players[i].index, game.Players[i].Player, game.Players[i].dead);
                            //printf("Player %d: %f %f %f | ROT %f | Local: %d\n", game.Players[i].index, game.Players[i].pos.x, game.Players[i].pos.y, game.Players[i].pos.z, game.Players[i].rot, game.Players[i].local);
                            //printf("localIndex %d\n", localPlayer.index);
                            //printf("name: %s\n", game.Players[i].name);
                            //std::string tmp = game.Players[i].name;

                            // This helps with dead ghosts, although the dead_3 offset i found works nicely i think
                            if (game.Players[i].dead > 0) {
                                game.Players[i].x = -30;
                                game.Players[i].y = 0;
                                game.Players[i].line = { 0, 0 };
                                continue;
                            }

                            // Position
                            n_x = (game.Players[i].pos.x - mapInfo.min_x) / (mapInfo.max_x - mapInfo.min_x);
                            n_y = (game.Players[i].pos.y - mapInfo.min_y) / (mapInfo.max_y - mapInfo.min_y);
                            game.Players[i].x = _Render.left + (n_x * (_Render.right - _Render.left));
                            game.Players[i].y = _Render.top + ((1 - n_y) * (_Render.bottom -  _Render.top));

                            // Rotation
                            rad = game.Players[i].rot * (pi / 180.0f);
                            game.Players[i].line.x = game.Players[i].x + (line_len + 200) * cos(rad);
                            game.Players[i].line.y = game.Players[i].y - (line_len + 200) * sin(rad);

                        }

                    }

                    std::this_thread::sleep_for(std::chrono::milliseconds(1));
                }

            }

        }

        printf("not in game\n");
        std::this_thread::sleep_for(std::chrono::milliseconds(5000));
    }

    return 0;
}
