#include "blackops6.h"

void Game::scatter_ingame(VMMDLL_SCATTER_HANDLE handle) {
    hand = handle;
    AddScatterRead(handle, Addresses::base + xbox::game_mode, &InGame, sizeof(Game::InGame));
}

bool Game::checkValid(VMMDLL_SCATTER_HANDLE handle) {
    valid = 0;

    for (int i = 0; i < Players.size(); i++) {
        if (Players[i].valid) {
            AddScatterRead(handle, Players[i].Player + xbox::player::valid, &Players[i].valid, sizeof(Players[i].valid));
        }
    }
    ExecuteScatterRead(handle);

    for (int i = 0; i < Players.size(); i++) {
        if (Players[i].valid) {
            valid++;
        }
    }

    if (valid != old_valid) {
        old_valid = valid;
        return false;
    }

    return true;
}

void Game::gatherPlayers(VMMDLL_SCATTER_HANDLE handle) {
    bool result;

    // Get player infos
    for (int i = 0; i < Players.size(); i++) {
        Players[i].info(handle);
    }
    result = ExecuteScatterRead(handle);
    if (!result) {
        printf("[!] gatherPlayers info failure\n");
    }

    // get name
    for (int i = 0; i < Players.size(); i++) {
        if (Players[i].valid) {
            AddScatterRead(handle, name_start + (xbox::name_array_size * Players[i].index + 4), &Players[i].name, sizeof(Players[i].name));
            old_valid += 1;
        }
    }
    result = ExecuteScatterRead(handle);
    if (!result) {
        printf("[!] gatherPlayers name failure\n");
    }
}

void Game::populatePlayers(VMMDLL_SCATTER_HANDLE handle, Player& localPlayer) {
    for (int i = 0; i < InGame; i++) {
        auto playerExists = [this](int index) {
            return std::any_of(Players.begin(), Players.end(), [index](const Player& player) {
                return player.index == index;
                });
            };

        if (playerExists(i)) {
            continue;
        }

        Player player;
        player.index = i;
        player.Player = (clientBase + (i * xbox::player::size));
   
        if (player.index == localPlayer.index) {
            player.local = true;
        }
        else {
            player.local = false;
        }
        Players.push_back(player);
    }
}

void Game::getBones(VMMDLL_SCATTER_HANDLE handle) {
    for (int i = 0; i < Players.size(); i++) {
        if (Players[i].valid) {
            uint16_t boneIndex = get_bone_index(Players[i].index);
            Players[i].bonePtr = ReadValue<uint64_t>(boneBase + (boneIndex * xbox::bone::size) + xbox::bone::offset);
        }
    }
}

void Player::get_index(uint64_t client_info) {
    uint64_t index_ptr = ReadPtr(client_info, xbox::local_index);
    Player::index = ReadValue<int>(index_ptr + xbox::local_index_pos);
}

void Player::info(VMMDLL_SCATTER_HANDLE handle) {
    AddScatterRead(handle, Player::Player + xbox::player::valid, &valid, sizeof(Player::valid));
    AddScatterRead(handle, Player::Player + xbox::player::team, &teamId, sizeof(Player::teamId));
    AddScatterRead(handle, Player::Player + xbox::player::pos, &posPtr, sizeof(Player::posPtr));
}

void Player::updatePlayer(VMMDLL_SCATTER_HANDLE handle) {
    AddScatterRead(handle, Player::posPtr + 0x50, &pos, sizeof(Player::pos));
    AddScatterRead(handle, Player::posPtr + 0x68, &rot, sizeof(Player::rot));
    AddScatterRead(handle, Player::Player + xbox::player::dead_3, &dead, sizeof(Player::dead));
}

void Player::getCam(VMMDLL_SCATTER_HANDLE handle) {
    AddScatterRead(handle, Addresses::base + xbox::camera_base, &camPtr, sizeof(Player::camPtr));
}

void Player::updateCam(VMMDLL_SCATTER_HANDLE handle) {
    AddScatterRead(handle, Player::camPtr + xbox::camera_pos, &cam, sizeof(Player::cam));
}

void Player::updateBonePtr(VMMDLL_SCATTER_HANDLE handle, uint64_t boneBase) {
    uint16_t boneIndex = get_bone_index(Player::index);
    Player::bonePtr = ReadValue<uint64_t>(boneBase + (boneIndex * xbox::bone::size) + xbox::bone::offset);    
}
