#include "../utils/Helper.h"
#include "../blops6/blackops6.h"
#include <thread>

void Render(Game& game) {
    MSG m;
    DXOverlay Overlay;
    DXRender Render = Overlay.CreateRender();
    int x1, x2;
    uint64_t color = 0xFFFF0000;

    do {
        if (PeekMessage(&m, Overlay.GetOverlay(), NULL, NULL, PM_REMOVE)) {
            TranslateMessage(&m);
            DispatchMessage(&m);
        }

        Render.BeginRender();

        // FPS (not game)
        //Render.DrawTextX(std::to_string(Render.GetFPS()), 10, 2, 0xffff0000, true, true);
    
        // Crosshair (green cross)
        //Render.DrawLine(ScreenCenterX - 7, ScreenCenterY, ScreenCenterX + 7, ScreenCenterY, 0xFF00FF00);
        //Render.DrawLine(ScreenCenterX, ScreenCenterY + 7, ScreenCenterX, ScreenCenterY - 7, 0xFF00FF00);

        if (game.Players.size() > 0) {

            for (int i = 0; i < game.Players.size(); i++) {

                if (game.Players[i].w2s.x > 1 && game.Players[i].w2s.y > 1) {
                    
                    std::lock_guard<std::mutex> lock(mutexlock);

                    if (game.Players[i].distance == 0 || game.Players[i].local) {
                        continue;
                    }
                    
                    x1 = game.Players[i].w2s.x + 200.0f / game.Players[i].distance;
                    x2 = game.Players[i].w2s.x - 200.0f / game.Players[i].distance;;
                    
                    // Bottom line
                    Render.DrawLine(game.Players[i].w2s.x, game.Players[i].w2s.y, x1, game.Players[i].w2s.y, color);
                    Render.DrawLine(game.Players[i].w2s.x, game.Players[i].w2s.y, x2, game.Players[i].w2s.y, color);

                    // Side lines
                    Render.DrawLine(x1, game.Players[i].w2s.y, x1, game.Players[i].head2d.y, color);
                    Render.DrawLine(x2, game.Players[i].w2s.y, x2, game.Players[i].head2d.y, color);

                    // Top line
                    Render.DrawLine(game.Players[i].w2s.x, game.Players[i].head2d.y, x1, game.Players[i].head2d.y, color);
                    Render.DrawLine(game.Players[i].w2s.x, game.Players[i].head2d.y, x2, game.Players[i].head2d.y, color);

                    // Name
                    Render.DrawTextX(game.Players[i].name, game.Players[i].w2s.x, game.Players[i].w2s.y, 0xFFFFFFEE);
                }
            }
        }

        Render.EndRender();

    } while (m.message != WM_QUIT);
}

bool world_to_screen(Vector3 world_location, Vector2& out, Vector3 camera_pos, int screen_width, int screen_height, Vector2 fov, Vector3 matricies[3]) {
    // Local Calculation
    auto local = world_location - camera_pos;

    // Transformation Calculation
    auto trans = Vector3{
        local.Dot(matricies[1]),
        local.Dot(matricies[2]),
        local.Dot(matricies[0])
    };

    // Check trans.z
    if (trans.z < 0.01f) {
        return false;
    }

    // Screen Coordinates Calculation
    out.x = ((float)screen_width / 2.0) * (1.0 - (trans.x / fov.x / trans.z));
    out.y = ((float)screen_height / 2.0) * (1.0 - (trans.y / fov.y / trans.z));

    // Bounds Check
    if (out.x < 1 || out.y < 1 || (out.x > decrypt_refdef->ref_def_nn.width) || (out.y > decrypt_refdef->ref_def_nn.height)) {
        return false;
    }

    return true;
}

auto w2s(Vector3 world_position, Vector2& screen_position, Vector3 cam_position) -> bool
{
    return world_to_screen(world_position, screen_position, cam_position, 1920, 1080, decrypt_refdef->ref_def_nn.tan_half_fov, decrypt_refdef->ref_def_nn.axis);
}

int main() {
	DMAController DMA;
	DMA.GetProcessInfo((LPSTR)"cod.exe");

    VMMDLL_SCATTER_HANDLE handle = CreateScatterHandle();
    Game game;
    Player localPlayer;
    bool result;

    std::thread RenderThread(Render, std::ref(game));
    while (true) {

        game.scatter_ingame(handle);
        result = ExecuteScatterRead(handle);
        if (!result) {
            printf("scatter read failed\n");
        }

        if (game.InGame > 1) {

            std::this_thread::sleep_for(std::chrono::milliseconds(7000)); // let it marinate
            game.clientInfo = decrypt_client_info();
            printf("in game!\n");

            if (game.clientInfo) {

                // decrypts
                game.clientBase = decrypt_client_base(game.clientInfo);
                game.boneBase = decrypt_bone_base();
                game.refDefPtr = decrypt_refdef->retrieve_ref_def();
                game.name_ptr = ReadValue<uint64_t>(Addresses::base + xbox::name_array);
                game.name_start = game.name_ptr + xbox::name_array_pos;

                // localPlayer setup and refDef
                localPlayer.get_index(game.clientInfo);
                localPlayer.Player = game.clientBase + (localPlayer.index * xbox::player::size);
                localPlayer.info(handle);
                localPlayer.getCam(handle);
                AddScatterRead(handle, game.refDefPtr, &decrypt_refdef->ref_def_nn, sizeof(ref_def_t));
                result = ExecuteScatterRead(handle);
                if (!result) {
                    printf("scatterread failed 1\n");
                }

                ref_def_t& ref_def = decrypt_refdef->ref_def_nn;
                game.populatePlayers(handle, localPlayer);
                game.gatherPlayers(handle);
                game.getBones(handle);

                int count = 0;
                while (true) {
                    count++;

                    // Update localplayer
                    localPlayer.updatePlayer(handle);
                    localPlayer.updateCam(handle);
                    AddScatterRead(handle, game.refDefPtr, &decrypt_refdef->ref_def_nn, sizeof(ref_def_t));
                    result = ExecuteScatterRead(handle);
                    if (!result) {
                        printf("localPlayer\n");
                    }

                    // extremely robust and mature leaver check
                    if (count == 1000) {
                        if (!game.checkValid(handle)) {
                            game.gatherPlayers(handle);
                        }
                        count = 0;
                    }

                    // Update player position
                    for (int i = 0; i < game.Players.size(); i++) {
                        if (game.Players[i].valid) {
                            game.Players[i].updatePlayer(handle);
                        }
                    }
                    
                    // Check if in game
                    game.scatter_ingame(handle);
                    result = ExecuteScatterRead(handle);
                    if (!result) {
                        printf("scatteread failed! 3\n");
                    }
                    if (game.InGame == 1) {
                        game.Players.clear();
                        break;
                    }

                    // bones
                    for (int i = 0; i < game.Players.size(); i++) {
                        if (game.Players[i].valid) {
                            AddScatterRead(handle, game.Players[i].bonePtr + (7 * 0x20) + 0x10, &game.Players[i].head_pos, sizeof(Player::head_pos));
                        }
                    }
                    ExecuteScatterRead(handle);
                    for (int i = 0; i < game.Players.size(); i++) {
                        if (game.Players[i].valid) {
                            game.Players[i].head3d = game.Players[i].pos + game.Players[i].head_pos;
                            game.Players[i].head3d.z += 60; // :(
                        }
                    }

                    // test update bonePtr
                    float l = 10.0f;
                    for (int i = 0; i < game.Players.size(); i++) {
                        if (game.Players[i].valid) {
                            if (game.Players[i].head3d.x > (game.Players[i].pos.x + l) || game.Players[i].head3d.x < (game.Players[i].pos.x - l)) {
                                game.Players[i].updateBonePtr(handle, game.boneBase);
                            }
                        }
                    }

                    // player loop
                    for (int i = 0; i < game.Players.size(); i++) {
                        if (game.Players[i].valid) {

                            // this might help with ghost boxes
                            if (game.Players[i].dead > 0 || game.Players[i].teamId == localPlayer.teamId || game.Players[i].local) {
                                game.Players[i].w2s = { 0,0 };
                                game.Players[i].head2d = { 0,0 };
                                continue;
                            }

                            std::lock_guard<std::mutex> lock(mutexlock);

                            // bones
                            w2s(game.Players[i].head3d, game.Players[i].head2d, localPlayer.cam);

                            // not bones
                            game.Players[i].distance = units_to_m(localPlayer.pos.DistTo(game.Players[i].pos));
                            w2s(game.Players[i].pos, game.Players[i].w2s, localPlayer.cam);
                        }
                    }
                    std::this_thread::sleep_for(std::chrono::milliseconds(1));
                }
            }
        }
        printf("not in game\n");
        std::this_thread::sleep_for(std::chrono::milliseconds(5000));
    }
    return 0;
}
