#include "MemoryController.h"
#include "../blops6/Addresses.h"

// constants
DWORD pid = 0;
static VMM_HANDLE vmmdll = 0;
bool Result;

// extra funcs
std::wstring pPath() {
    TCHAR buffer[MAX_PATH] = { 0 };
    GetModuleFileName(NULL, buffer, MAX_PATH);
    std::wstring path(buffer);
    std::wstring::size_type pos = path.find_last_of(L"\\/");
    return path.substr(0, pos);
}

bool MmapExists(const std::wstring& name) {
    struct _stat buffer;
    return (_wstat(name.c_str(), &buffer) == 0);
}

LPSTR ConvertWStringToLPSTR(const std::wstring& wstr) {
    int size_needed = WideCharToMultiByte(CP_ACP, 0, wstr.c_str(), (int)wstr.size(), NULL, 0, NULL, NULL);
    LPSTR strTo = new char[size_needed + 1];
    WideCharToMultiByte(CP_ACP, 0, wstr.c_str(), (int)wstr.size(), strTo, size_needed, NULL, NULL);
    strTo[size_needed] = '\0';
    return strTo;
}

// =====================================
// = DMA
// =====================================
DMAController::DMAController() {
    bool result = DMAController::Initialize();
    if (!result) {
        exit(0);
    }
}

DMAController::~DMAController() {
    DMAController::Close();
}

bool DMAController::Initialize() {
    std::wstring basePath = pPath();
    std::wstring suffix = L"\\mmap.txt";
    std::wstring filePath = basePath + suffix;

    if (MmapExists(filePath)) {
        printf("[+] mmap.txt found!\n");
        LPSTR args[] = { (LPSTR)"", (LPSTR)"-device", (LPSTR)"fpga", (LPSTR)"-memmap", ConvertWStringToLPSTR(filePath) };
        DWORD argc = 5;
        vmmdll = VMMDLL_Initialize(argc, args);
        printf("[+] VMMDLL_Initialize complete!\n");
        return true;
    }

    printf("[!] No mmap.txt found... generating. Restart program after!\n");
    LPSTR args[] = { (LPSTR)"", (LPSTR)"-device", (LPSTR)"fpga", (LPSTR)"-v" };
    DWORD argc = 4;

    vmmdll = VMMDLL_Initialize(argc, args);
    if (vmmdll) {
        printf("[+] VMMDLL_Initialize()\n");
        PVMMDLL_MAP_PHYSMEM pMPhys = NULL;
        auto res = VMMDLL_Map_GetPhysMem(vmmdll, &pMPhys);
        if (res) {
            if (pMPhys->dwVersion != VMMDLL_MAP_PHYSMEM_VERSION) {
                printf("[!] Failure: VMMDLL_MAP_PHYSMEM_VERSION\n");
                VMMDLL_MemFree(pMPhys);
                VMMDLL_Close(vmmdll);
                return false;
            }

            // Credit Metick@github.com, or whoever made this, stole your code
            std::stringstream sb;
            for (DWORD i = 0; i < pMPhys->cMap; i++) {
                sb << std::setfill('0') << std::setw(4) << i << "  " << std::hex << pMPhys->pMap[i].pa << "  -  " << (pMPhys->pMap[i].pa + pMPhys->pMap[i].cb - 1) << "  ->  " << pMPhys->pMap[i].pa << std::endl;
            }
            std::ofstream nFile("mmap.txt");
            nFile << sb.str();
            nFile.close();

            printf("[+] Dumped mmap.txt\n");
            VMMDLL_MemFree(pMPhys);
            printf("[+] Memory free'd\n");
            DMAController::Close();
            return false;
        }
    }
    printf("[!] Failed: VMMDLL_Initialize()\n");
    return false;
}

bool DMAController::GetProcessInfo(LPSTR processName) {
    if (!VMMDLL_PidGetFromName(vmmdll, processName, &pid)) {
        printf("[!] Failed to find %s\n", processName);
        return false;
    }
    else {
        printf("[+] Found %s PID: %d\n", processName, pid);
        VMMDLL_PROCESS_INFORMATION pProcInfo;
        SIZE_T procInfoSz = sizeof(pProcInfo);
        //ZeroMemory(&ProcessInformation, sizeof(VMMDLL_PROCESS_INFORMATION));
        pProcInfo.magic = VMMDLL_PROCESS_INFORMATION_MAGIC;
        pProcInfo.wVersion = VMMDLL_PROCESS_INFORMATION_VERSION;
        Result = VMMDLL_ProcessGetInformation(vmmdll, pid, &pProcInfo, &procInfoSz);

        if (Result) {
            printf("[+] Success: VMMDLL_ProcessGetInformation()\n\n");
            printf("Name: %s\n", pProcInfo.szNameLong);
            printf("PID: %d\n", pProcInfo.dwPID);
            printf("PPID: %d\n", pProcInfo.dwPPID);
            printf("ProcessState: %d\n", pProcInfo.dwState);
            printf("\n");
        }
        else {
            printf("[!] Fail: VMMDLL_ProcessGetInformation()\n");
            return false;
        }

        ULONG64 processBase;
        PVMMDLL_MAP_MODULEENTRY pModuleEntryExplorer;
        Result = VMMDLL_Map_GetModuleFromNameU(vmmdll, pid, processName, &pModuleEntryExplorer, NULL);
        if (!Result) {
            printf("[!] Failed: VMMDLL_Map_GetModuleFromNameW\n");
            return false;
        }

        printf("[+] Base: 0x%llx\n", pModuleEntryExplorer->vaBase);
        printf("[+] Size: 0x%lx\n", pModuleEntryExplorer->cbImageSize);
        printf("[+] End: 0x%llx\n", pModuleEntryExplorer->vaBase + pModuleEntryExplorer->cbImageSize);

        Addresses addr;
        uint64_t base = pModuleEntryExplorer->vaBase;

        addr.SetAddress(base);

        return true;
    }
}

void DMAController::Close() {
    VMMDLL_Close(vmmdll);
    printf("[+] VMMDLL_Close()\n");
    return;
}

// =====================================
// = READ/WRITE FUNCTIONS
// =====================================

BYTE* ReadMemory(uint64_t address, DWORD count) {
    DWORD cbRead;
    BYTE pb[0x1000];

    if (!VMMDLL_MemReadEx(vmmdll, pid, address, pb, count, &cbRead, 0)) {
        printf("[!] FAIL: VMMDLL_MemReadEx\n");
        return NULL;
    }

    if (count != cbRead) {
        BYTE* npb = new BYTE[cbRead];
        memcpy(npb, pb, cbRead);
        return npb;
    }
    return pb;
}

uintptr_t ReadPtr(uint64_t ptr, uint64_t offset) {
    uintptr_t address = ReadValue<uintptr_t>(ptr + offset);
    if (address != 0x0) {
        return address;
    }
    return 0x0;
}

uintptr_t ReadPtrChain(uint64_t ptr, std::vector<uint64_t>& offsets) {
    uintptr_t address = 0;
    address = ReadPtr(ptr, offsets[0]);
    
    if (address == 0x0) {
        printf("[!] ReadPtrChain bad first offset\n");
    }

    for (int i = 1; i < offsets.size(); i++) {
        address = ReadPtr(address, offsets[i]);
        if (address == 0x0) {
            printf("[!] ReadPtrChain bad %d offset\n", i);
        }
    }

    return address;
}

template <typename T>
T ReadValue(uint64_t address) {
    size_t size = sizeof(T);
    T value;
    BYTE* buffer = new BYTE[size];

    if (!VMMDLL_MemRead(vmmdll, pid, address, buffer, size)) {
        //printf("[!] ReadValue Failure\n");
    }

    memcpy(&value, buffer, size);
    delete[] buffer;
    return value;
}

uint64_t GetPEBAddress() {
    VMMDLL_PROCESS_INFORMATION procInfo = { 0 };
    SIZE_T cbProcInfo = sizeof(VMMDLL_PROCESS_INFORMATION);

    procInfo.magic = VMMDLL_PROCESS_INFORMATION_MAGIC;
    procInfo.wVersion = VMMDLL_PROCESS_INFORMATION_VERSION;
    procInfo.wSize = (WORD)cbProcInfo;

    if (!VMMDLL_ProcessGetInformation(vmmdll, pid, &procInfo, &cbProcInfo)) {
        std::cerr << "[!] Failed to get process information for PID: " << pid << std::endl;
        return 0x0;
    }

    uint64_t PebAddress = procInfo.win.vaPEB;
    return PebAddress;
}

// =====================================
// = SCATTER FUNCTIONS                 =
// =====================================
VMMDLL_SCATTER_HANDLE CreateScatterHandle() {
    VMMDLL_SCATTER_HANDLE VMMDLL_Scatter_Handle = VMMDLL_Scatter_Initialize(vmmdll, pid, VMMDLL_FLAG_NOCACHE);
    return VMMDLL_Scatter_Handle;
}

void AddScatterRead(VMMDLL_SCATTER_HANDLE hS, uint64_t address, void* buffer, size_t size ) {
    bool result = VMMDLL_Scatter_PrepareEx(hS, address, size, (PBYTE)buffer, NULL); 
    if (!result) {
        printf("[!] Failed to add scatterread entry for %llx\n", address);
    }
}

bool ExecuteScatterRead(VMMDLL_SCATTER_HANDLE hS) {
    bool result = VMMDLL_Scatter_ExecuteRead(hS);
    if (!result) {
        return false;
    }

    result = VMMDLL_Scatter_Clear(hS, pid, VMMDLL_FLAG_NOCACHE);
    if (!result) {
        return false;
    }

    return true;
}

void CloseScatterHandle(VMMDLL_SCATTER_HANDLE VMMDLL_Scatter_Handle) {
    VMMDLL_Scatter_CloseHandle(VMMDLL_Scatter_Handle);
}

// =====================================
// = Templates for ReadValue
// =====================================
template Vector3 ReadValue<Vector3>(uint64_t address);
template Vector2 ReadValue<Vector2>(uint64_t address);

template uint8_t ReadValue<uint8_t>(uint64_t address);
template uint16_t ReadValue<uint16_t>(uint64_t address);
template uint32_t ReadValue<uint32_t>(uint64_t address);
template uint64_t ReadValue(uint64_t address);

template int ReadValue<int>(uint64_t Address);
template float ReadValue<float>(uint64_t Address);
template bool ReadValue<bool>(uint64_t Address);

template DWORD ReadValue<DWORD>(uint64_t address);
template std::string ReadValue<std::string>(uint64_t address);
template FCameraCacheEntry ReadValue<FCameraCacheEntry>(uint64_t address);

// custom codshit
template key ReadValue<key>(uint64_t Address);
template ref_def_t ReadValue<ref_def_t>(uint64_t Address);