#include <ctime>
#include <Windows.h>
#include "Globals.h"
#include "Overlay.h"
#include "Render.h"

std::string GetString(const int len) {
    srand(static_cast<unsigned int>(time(0)));
    static const char alphanum[] =
        "0123456789"
        "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz";
    std::string tmp_s;
    tmp_s.reserve(len);

    for (int i = 0; i < len; ++i) {
        tmp_s += alphanum[rand() % (sizeof(alphanum) - 1)];
    }

    return tmp_s;
}

std::wstring Convert(const std::string& as) {
    if (as.empty())    return std::wstring();
    size_t reqLength = ::MultiByteToWideChar(CP_UTF8, 0, as.c_str(), (int)as.length(), 0, 0);
    std::wstring ret(reqLength, L'\0');
    ::MultiByteToWideChar(CP_UTF8, 0, as.c_str(), (int)as.length(), &ret[0], (int)ret.length());
    return ret;
}

std::string X = GetString(16);
std::wstring Y = Convert(X);
const wchar_t* cNAME = Y.c_str();
const wchar_t* wTITLE = Y.c_str();
LPCWSTR CLASS_NAME = (LPCWSTR)cNAME;
LPCWSTR WINDOW_TITLE = (LPCWSTR)wTITLE;

DXOverlay::DXOverlay() {
    CreateOverlay();
}

DXRender DXOverlay::CreateRender() {
    return DXRender(vDevice);
}

DXOverlay::~DXOverlay() {
    if (OverlayWindow) {
        DestroyWindow(OverlayWindow);
    }

    if (D3D) {
        D3D->Release();
    }

    if (vDevice) {
        vDevice->Release();
    }
}

HWND DXOverlay::GetOverlay() {
    return OverlayWindow;
}

bool DXOverlay::IsInit() {
    return sInit;
}

void DXOverlay::CreateOverlay() {
    WNDCLASS wc = { };
    HINSTANCE hInstance = GetModuleHandle(NULL);

    wc.style = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc = WndProc;
    wc.hInstance = hInstance;
    wc.lpszClassName = CLASS_NAME;
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);

    if (!RegisterClass(&wc)) {
        MessageBox(NULL, L"Window class registration failed!", L"Error", MB_ICONERROR);
        return;
    }

    OverlayWindow = CreateWindowEx(
        WS_EX_LAYERED | WS_EX_TRANSPARENT, // | WS_EX_TOPMOST,
        CLASS_NAME,                    // Window class
        WINDOW_TITLE,                  // Window title
        WS_POPUP,//WS_OVERLAPPEDWINDOW,          // Window style
        0, 0,
        GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN),
        NULL,                          // Parent window
        NULL,                          // Menu
        hInstance,                     // Instance handle
        NULL                           // Additional application data
    );

    if (OverlayWindow == NULL) {
        MessageBox(NULL, L"Window creation failed!", L"Error", MB_ICONERROR);
        return;
    }

    COLORREF wColor = RGB(255, 255, 255);
    SetLayeredWindowAttributes(OverlayWindow, wColor, 0, LWA_COLORKEY);
    ShowWindow(OverlayWindow, SW_SHOWDEFAULT);

    InitDX9();
}

void DXOverlay::InitDX9() {
    D3D = Direct3DCreate9(D3D_SDK_VERSION);

    if (!D3D)
        throw std::exception("[!] InitDX9: failed to create object\n");

    std::cout << "[+] DirectX Obj created!\n";

    D3DPRESENT_PARAMETERS d3d_pp;
    ZeroMemory(&d3d_pp, sizeof(d3d_pp));

    d3d_pp.Windowed = true;
    d3d_pp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3d_pp.BackBufferFormat = D3DFMT_A8R8G8B8;
    d3d_pp.BackBufferWidth = GetSystemMetrics(SM_CXSCREEN);
    d3d_pp.BackBufferHeight = GetSystemMetrics(SM_CYSCREEN);
    d3d_pp.hDeviceWindow = OverlayWindow;
    d3d_pp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;

    if (FAILED(D3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, OverlayWindow, D3DCREATE_HARDWARE_VERTEXPROCESSING, &d3d_pp, &vDevice))) {
        D3D->Release();
        throw std::exception("[!] Failed to create device");
    }

    std::cout << "[+] DirectX device created!\n";

    sInit = true;
}

LRESULT CALLBACK DXOverlay::WndProc(HWND wnd, UINT msg, WPARAM wparam, LPARAM lparam) {
    switch (msg) {
    case WM_DESTROY:
        exit(0);
        break;
    default:
        return DefWindowProc(wnd, msg, wparam, lparam);
    }
}