#include "Render.h"
#include "Helper.h"

DXRender::DXRender(IDirect3DDevice9* Device) {
	if (!Device) {
		printf("[!] Invalid device ptr, NULL!\n");
	}

	xDevice = Device;

	if (FAILED(D3DXCreateLine(xDevice, &Line))) {
		std::cout << "[!] Failed to CreateLine!\n";
		throw std::exception("[!] Failed to createline\n");
	}

	if (FAILED(D3DXCreateFont(xDevice, 9, NULL, FW_HEAVY, 1, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, ANTIALIASED_QUALITY, DEFAULT_PITCH | FF_DONTCARE, (LPWSTR)"Tahoma", &Font))) {
		std::cout << "[!] Failed to CreateFont!\n";
		throw std::exception("[!] Failed to create font\n");
	}
}

DXRender::~DXRender() {
	if (Font) {
		Font->Release();
	}
	if (Line) {
		Line->Release();
	}
}

void DXRender::BeginRender() {
	xDevice->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_ARGB(0, 0, 0, 0), 1.0f, 0);
	xDevice->BeginScene();
}

void DXRender::EndRender() {
	static int FPS = 0;
	static float ltc = 0.f;

	FPS++;

	float ctc = clock() * 0.001f;

	if (ctc - ltc >= 1.f) {
		ltc = ctc;
		sFPS = FPS;
		FPS = 0;
	}

	xDevice->EndScene();
	xDevice->Present(NULL, NULL, NULL, NULL);
}

int DXRender::GetFPS() {
	return sFPS;
}

void DXRender::DrawRectangle(int x, int y, int w, int h, unsigned long color) {
	DrawLine(x, y, x + w, y, color);
	DrawLine(x, y, x, y + h, color);
	DrawLine(x + w, y, x + w, y + h, color);
	DrawLine(x, y + h, x + w + 1, y + h, color);
}

void DXRender::DrawFilledRectangle(int x, int y, int w, int h, unsigned long color) {
	D3DTLVERTEX qV[4] = {
		{ float(x), float(y + h), 0.f, 1.f, color },
		{ float(x), float(y), 0.f, 1.f, color },
		{ float(x + w), float(y + h), 0.f, 1.f, color },
		{ float(x + w), float(y) , 0.f, 1.f, color }
	};

	xDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, true);
	xDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
	xDevice->SetFVF(D3DFVF_XYZRHW | D3DFVF_DIFFUSE | D3DFVF_TEX1);
	xDevice->SetTexture(0, nullptr);
	xDevice->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, qV, sizeof(D3DTLVERTEX));
}

void DXRender::DrawOutlinedRectangle(int x, int y, int w, int h, unsigned long color) {
	DrawRectangle(x - 1, y - 1, w + 2, h + 2, D3DCOLOR_RGBA(1, 1, 1, 255));
	DrawRectangle(x + 1, y + 1, w - 2, h - 2, D3DCOLOR_RGBA(1, 1, 1, 255));
	DrawRectangle(x, y, w, h, color);
}

void DXRender::DrawTextX(std::string text, int x, int y, unsigned long color, bool center, bool outline) {
	if (center) {
		RECT dimensions = GetTextDimensions(text);
		x -= (dimensions.right - dimensions.left) / 2;
	}

	auto _text = [&](std::string _text, int _x, int _y, unsigned long _color) {
		//std::cout << "[+] Drawing text " << _text << "\n";
		RECT r{ _x, _y, _x, _y };
		Font->DrawTextA(NULL, _text.data(), -1, &r, DT_NOCLIP, _color);
		};

	if (outline) {
		_text(text, x - 1, y, D3DCOLOR_RGBA(1, 1, 1, 255));
		_text(text, x + 1, y, D3DCOLOR_RGBA(1, 1, 1, 255));
		_text(text, x, y - 1, D3DCOLOR_RGBA(1, 1, 1, 255));
		_text(text, x, y + 1, D3DCOLOR_RGBA(1, 1, 1, 255));
	}

	_text(text, x, y, color);
}

RECT DXRender::GetTextDimensions(std::string text) {
	RECT r;
	Font->DrawTextA(NULL, text.data(), -1, &r, DT_CALCRECT, 0xFFFFFFFF);
	return r;
}

void DXRender::DrawLine(int x0, int y0, int x1, int y1, unsigned long color) {
	D3DXVECTOR2 lines[2] = {
		D3DXVECTOR2(x0, y0),
		D3DXVECTOR2(x1, y1)
	};

	Line->Begin();
	Line->Draw(lines, 2, color);
	Line->End();
}